from enum import Enum


class FrameType(Enum):
    RESPONSE = 0
    ERROR = 1
    MESSAGE = 2

    @property
    def is_response(self) -> bool:
        return self == self.RESPONSE

    @property
    def is_error(self) -> bool:
        return self == self.ERROR

    @property
    def is_message(self) -> bool:
        return self == self.MESSAGE



class NSQHTTPException(Exception):
    """XXX"""


class TransportError(NSQHTTPException):
    """XXX"""

    @property
    def status_code(self):
        """XXX"""
        return self.args[0]

    @property
    def error(self):
        """A string error message."""
        return self.args[1]

    @property
    def info(self):
        """Dict of returned error info from ES, where available."""
        return self.args[2]

    def __str__(self):
        return 'TransportError(%s, %r)' % (self.status_code, self.error)


class HTTPConnectionError(TransportError):
    """XXX"""

    def __str__(self):
        return 'HttpConnectionError(%s) caused by: %s(%s)' % (
            self.error, self.info.__class__.__name__, self.info)


class NotFoundError(TransportError):
    """Exception representing a 404 status code."""


class ConflictError(TransportError):
    """Exception representing a 409 status code."""


class RequestError(TransportError):
    """Exception representing a 400 status code."""


# more generic mappings from status_code to python exceptions
HTTP_EXCEPTIONS = {
    400: RequestError,
    404: NotFoundError,
}



import json

import aiohttp

from .http_exceptions import HTTP_EXCEPTIONS, NSQHTTPException
from ..utils import convert_to_str


class NSQHTTPConnection:
    """XXX"""

    def __init__(self, host='127.0.0.1', port=4150, *, loop):
        self._loop = loop
        self._endpoint = (host, port)
        self._base_url = 'http://{0}:{1}/'.format(*self._endpoint)

        self._session = aiohttp.ClientSession()

    @property
    def endpoint(self):
        return 'http://{0}:{1}'.format(*self._endpoint)

    async def close(self):
        return await self._session.close()

    async def perform_request(self, method, url, params, body):
        _body = convert_to_str(body) if body else body
        url = self._base_url + url
        print(method, url, params, _body)
        try:
            resp = await self._session.request(method, url,
                                               params=params,
                                               data=_body)
        except Exception as tmp:
            print('exception', tmp)
        resp_body = await resp.text()
        try:
            response = json.loads(resp_body)
        except ValueError:
            return resp_body

        if not (200 <= resp.status <= 300):
            extra = None
            try:
                extra = json.loads(resp_body)
            except ValueError:
                pass
            exc_class = HTTP_EXCEPTIONS.get(resp.status, NSQHTTPException)
            raise exc_class(resp.status, resp_body, extra)
        return response

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '<{}: {}>'.format(cls_name, self._endpoint)



"""NSQ protocol parser.

:see: https://nsq.io/clients/tcp_protocol_spec.html
"""
import abc
import struct
from typing import Any, Optional, Tuple, Union

from ansq.tcp import consts
from ansq.tcp.exceptions import ProtocolError
from ansq.tcp.types import (
    FrameType, NSQResponseSchema, NSQMessageSchema, NSQErrorSchema
)
from ansq.utils import convert_to_bytes

__all__ = 'Reader'


class BaseReader(metaclass=abc.ABCMeta):
    @abc.abstractmethod  # pragma: no cover
    def feed(self, chunk):
        pass

    @abc.abstractmethod  # pragma: no cover
    def get(self):
        pass

    @abc.abstractmethod  # pragma: no cover
    def encode_command(self, cmd, *args, data=None):
        pass


class Reader(BaseReader):
    def __init__(self, buffer: bytes = None):
        self._buffer = bytearray()
        self._is_header = False
        self._payload_size = 0
        buffer and self.feed(buffer)

    @property
    def buffer(self) -> bytearray:
        return self._buffer

    def feed(self, chunk: bytes):
        """Put raw chunk of data obtained from connection to buffer.

        :param chunk: Raw input data.
        :type chunk: :class:`bytes`
        """
        chunk and self._buffer.extend(chunk)

    def get(self) -> Optional[
        Union[NSQResponseSchema, NSQErrorSchema, NSQMessageSchema]
    ]:
        """Get from buffer NSQ response

        :raises ProtocolError: On unexpected NSQ message's FrameType
        :returns: Depends of ``frame_type``, returns
            :class:`NSQResponse`, :class:`NSQError`,  or :class:`NSQMessage`
        """
        buffer_size = len(self._buffer)

        if not self._is_header and buffer_size >= consts.DATA_SIZE:
            size = struct.unpack('>l', self._buffer[:consts.DATA_SIZE])[0]
            self._payload_size = size
            self._is_header = True

        if (
                self._is_header
                and buffer_size >= consts.DATA_SIZE + self._payload_size
        ):
            start, end = consts.DATA_SIZE, consts.HEADER_SIZE
            frame_type = FrameType(
                struct.unpack('>l', self._buffer[start:end])[0])
            resp = self._parse_payload(frame_type, self._payload_size)

            self._buffer = self._buffer[start + self._payload_size:]
            self._is_header = False
            self._payload_size = 0

            return resp

    def _parse_payload(
            self, frame_type: FrameType, payload_size: int
    ) -> Union[NSQResponseSchema, NSQErrorSchema, NSQMessageSchema]:
        """Parse from buffer NSQ response

        :raises ProtocolError: On unexpected NSQ message's FrameType
        :returns: Depends of ``frame_type``, returns
            :class:`NSQResponse`, :class:`NSQError`,  or :class:`NSQMessage`
        """
        if frame_type == FrameType.RESPONSE:
            return NSQResponseSchema(
                self._unpack_response(payload_size), frame_type=frame_type)
        if frame_type == FrameType.ERROR:
            return NSQErrorSchema(
                *self._unpack_error(payload_size), frame_type=frame_type)
        if frame_type == FrameType.MESSAGE:
            return NSQMessageSchema(
                *self._unpack_message(payload_size), frame_type=frame_type)

        raise ProtocolError('Got unexpected FrameType: {}'.format(frame_type))

    def _unpack_response(self, payload_size: int) -> bytes:
        """Unpack the response from the buffer"""
        start = consts.HEADER_SIZE
        end = consts.DATA_SIZE + payload_size
        return bytes(self._buffer[start:end])

    def _unpack_error(self, payload_size: int) -> Tuple[bytes, bytes]:
        """Unpack the error from the buffer"""
        error = self._unpack_response(payload_size)
        code, msg = error.split(maxsplit=1)
        return code, msg

    def _unpack_message(self, payload_size: int) -> Tuple[
            int, int, bytes, bytes]:
        """Unpack the message from the buffer.

        :see: https://docs.python.org/3/library/struct.html
        :rtype: :class:`NSQMessageSchema`
        :returns: NSQ Message
        """
        start = consts.HEADER_SIZE
        end = consts.DATA_SIZE + payload_size
        msg_len = end - start - consts.MSG_HEADER
        fmt = '>qh16s{}s'.format(msg_len)
        return struct.unpack(fmt, self._buffer[start:end])

    def encode_command(self, cmd: str, *args, data: Any = None) -> bytes:
        """Encode command to bytes"""
        _cmd = convert_to_bytes(cmd.upper().strip())
        _args = [convert_to_bytes(a) for a in args]
        body_data, params_data = b'', b''

        if len(_args):
            params_data = b' ' + b' '.join(_args)

        if data and isinstance(data, (list, tuple)):
            data_encoded = [self._encode_body(part) for part in data]
            num_parts = len(data_encoded)
            payload = struct.pack('>l', num_parts) + b''.join(data_encoded)
            body_data = struct.pack('>l', len(payload)) + payload
        elif data:
            body_data = self._encode_body(data)

        return b''.join((_cmd, params_data, consts.NL, body_data))

    @staticmethod
    def _encode_body(data: Any) -> bytes:
        _data = convert_to_bytes(data)
        result = struct.pack('>l', len(_data)) + _data
        return result



from enum import Enum


class ConnectionStatus(Enum):
    CLOSED = 0
    CLOSING = 0
    INIT = 1
    CONNECTED = 2
    SUBSCRIBED = 3
    RECONNECTING = 4

    @property
    def is_closed(self) -> bool:
        return self == self.CLOSED

    @property
    def is_closing(self) -> bool:
        return self == self.CLOSING

    @property
    def is_init(self) -> bool:
        return self == self.INIT

    @property
    def is_connected(self) -> bool:
        return self == self.CONNECTED

    @property
    def is_subscribed(self) -> bool:
        return self == self.SUBSCRIBED

    @property
    def is_reconnecting(self) -> bool:
        return self == self.RECONNECTING

    def __bool__(self):
        return not self.is_closed and not self.is_closing and not self.is_init



import abc
import asyncio
import logging
from asyncio.events import AbstractEventLoop
from asyncio.streams import StreamReader, StreamWriter
from collections import deque
from typing import Optional, Union, Callable, Any, TYPE_CHECKING

from ansq.tcp.protocol import Reader
from ansq.tcp.types import ConnectionStatus, NSQMessage, NSQCommands
from ansq.utils import get_logger

if TYPE_CHECKING:
    from ansq.tcp.types import (
        NSQResponseSchema, NSQErrorSchema, NSQMessageSchema
    )


class TCPConnection(abc.ABC):
    def __init__(
            self, host: str = 'localhost', port: int = 4150, *,
            message_queue: asyncio.Queue = None, on_message: Callable = None,
            on_exception: Callable = None, loop: AbstractEventLoop = None,
            auto_reconnect: bool = True, heartbeat_interval: int = 30000,
            feature_negotiation: bool = True, tls_v1: bool = False,
            snappy: bool = False, deflate: bool = False,
            deflate_level: int = 6, sample_rate: int = 0,
            debug: bool = False, logger: logging.Logger = None
    ):
        self._host, self._port = host, port
        self._loop: AbstractEventLoop = loop or asyncio.get_event_loop()
        self._debug = debug
        self.logger = logger or get_logger(debug)

        self._message_queue: asyncio.Queue[Optional[NSQMessage]] = \
            message_queue or asyncio.Queue()
        self._status: ConnectionStatus = ConnectionStatus.INIT
        self._reader: Optional[StreamReader] = None
        self._writer: Optional[StreamWriter] = None
        self._reader_task: Optional[asyncio.Task] = None
        self._reconnect_task: Optional[asyncio.Future] = None
        self._auto_reconnect = auto_reconnect
        self._parser = Reader()

        self._config = {
            'deflate': deflate,
            'deflate_level': deflate_level,
            'sample_rate': sample_rate,
            'snappy': snappy,
            'tls_v1': tls_v1,
            'heartbeat_interval': heartbeat_interval,
            'feature_negotiation': feature_negotiation,
        }

        self._last_message_timestamp: Optional[float] = None
        # Next queue is used for nsq commands
        self._cmd_waiters = deque()
        # Mark connection in upgrading state to ssl socket
        self._is_upgrading = False
        # Number of received but not acked or req messages
        self._in_flight = 0
        self._secret: Optional[str] = None
        self._is_auth_required = False
        self._is_authorized = False

        # Handlers
        self._on_message = on_message
        self._on_exception = on_exception

        # Reader setup
        self._topic: Optional[str] = None
        self._channel: Optional[str] = None
        self.rdy_messages_count: int = 1
        self._is_subscribed = False

    def __repr__(self) -> str:
        return '<{class_name}: endpoint={endpoint}, status={status}>'.format(
            class_name=self.__class__.__name__, endpoint=self.endpoint,
            status=self.status)

    @property
    def status(self) -> ConnectionStatus:
        return self._status

    @property
    def endpoint(self) -> str:
        return 'tcp://{}:{}'.format(self._host, self._port)

    @property
    def in_flight(self) -> int:
        return self._in_flight

    @property
    def message_queue(self) -> 'asyncio.Queue[NSQMessage]':
        return self._message_queue

    @property
    def last_message(self) -> float:
        return self._last_message_timestamp

    @property
    def is_subscribed(self) -> bool:
        return self._is_subscribed

    @property
    def subscribed_topic(self) -> Optional[str]:
        return self._topic

    @property
    def subscribed_channel(self) -> Optional[str]:
        return self._channel

    @property
    def is_auth_required(self) -> bool:
        return self._is_auth_required

    @property
    def is_authorized(self) -> bool:
        return self._is_authorized

    @property
    def is_closed(self) -> bool:
        """True if connection is closed or closing."""
        return self.status.is_closed or self._status.is_closing

    @abc.abstractmethod
    async def connect(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    async def reconnect(self):
        raise NotImplementedError()

    async def close(self):
        """Cleanly close your connection (no more messages are sent)"""
        await self._do_close()

    async def cls(self):
        """Alias command for ``close()``."""
        await self.close()

    @abc.abstractmethod
    async def _do_close(self, exception: Exception = None):
        raise NotImplementedError()

    @abc.abstractmethod
    async def execute(
            self, command: Union[str, bytes], *args, data: Any = None,
            callback: Callable = None
    ) -> Optional[
        Union['NSQResponseSchema', 'NSQErrorSchema', 'NSQMessageSchema']
    ]:
        raise NotImplementedError()

    @abc.abstractmethod
    async def identify(self, **config):
        raise NotImplementedError()

    async def _pulse(self) -> None:
        await self.execute(NSQCommands.NOP)

    @abc.abstractmethod
    async def _upgrade_to_tls(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def _upgrade_to_snappy(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def _upgrade_to_deflate(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def _read_data_task(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def _parse_data(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def _on_message_hook(self, response: 'NSQMessageSchema'):
        raise NotImplementedError()

    @abc.abstractmethod
    async def _read_buffer(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def _start_upgrading(self, resp=None):
        raise NotImplementedError()

    @abc.abstractmethod
    async def _finish_upgrading(self, resp=None):
        raise NotImplementedError()



import re
import logging
from decimal import Decimal
from functools import singledispatch
from typing import Union, Tuple, Optional
from urllib.parse import urlsplit


def get_host_port(uri: str) -> Tuple[str, Optional[int]]:
    """Get host and port from provided URI."""
    split_uri = urlsplit(uri)
    return split_uri.hostname, split_uri.port


def validate_topic_channel_name(name: str):
    """Validate topic/channel names.
    The regex is ``^[.a-zA-Z0-9_-]{2,64}+(#ephemeral)?$``

    :raises AssertionError: Value not matches regex.
    """
    assert re.match(r'^[.a-zA-Z0-9_\-]{2,64}(#ephemeral)?$', name), (
        'Topic name must matches ^[.a-zA-Z0-9_-]{2,64}+(#ephemeral)?$ regex')


@singledispatch
def convert_to_bytes(value):
    """Base dispatch for unregistered convertible types

    :raises TypeError:
    """
    raise TypeError(
        'Argument {} expected to be type of '
        'bytes, str, int or float'.format(value))


@convert_to_bytes.register(bytes)
@convert_to_bytes.register(bytearray)
def _(value: Union[bytes, bytearray]) -> bytes:
    """Convert ``bytes`` or ``bytearray`` to bytes"""
    return value


@convert_to_bytes.register(str)
def _(value: str) -> bytes:
    """Convert ``str`` to bytes"""
    return value.encode('utf-8')


@convert_to_bytes.register(int)
@convert_to_bytes.register(float)
@convert_to_bytes.register(dict)
@convert_to_bytes.register(Decimal)
def _(value: Union[int, float, Decimal]) -> bytes:
    """Convert ``int``, ``float``, ``dict`` or ``Decimal`` to bytes"""
    return str(value).encode('utf-8')


@singledispatch
def convert_to_str(value):
    """Base dispatch for unregistered convertible types

    :raises TypeError:
    """
    raise TypeError(
        'Argument {} expected to be type of '
        'bytes, str, int, float or dict'.format(value))


@convert_to_str.register(str)
def _(value: str) -> str:
    """Convert ``str`` to str"""
    return value


@convert_to_str.register(bytes)
def _(value: bytes) -> str:
    """Convert ``bytes`` to str"""
    return value.decode('utf-8')


@convert_to_str.register(bytearray)
def _(value: bytearray) -> str:
    """Convert ``bytearray`` to str"""
    return bytes(value).decode('utf-8')


@convert_to_str.register(int)
@convert_to_str.register(float)
def _(value: Union[int, float]) -> str:
    """Convert ``int`` or ``float`` to str"""
    return str(value)


def get_logger(debug: bool = False):
    """Get the ansq logger.

    :params debug: Set up debug level.
    :type debug: :class:`bool`
    """
    logger = logging.getLogger('ansq')
    log_format = "%(asctime)s - %(levelname)s - %(name)s: %(message)s"
    logging.basicConfig(format=log_format)
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    return logger



from .commands import NSQCommands
from .connection_status import ConnectionStatus
from .frame_type import FrameType
from .message import NSQMessage
from .response_schemas import (
    NSQResponseSchema, NSQMessageSchema, NSQErrorSchema
)
from .connection import TCPConnection

__all__ = (
    'TCPConnection', 'NSQMessage', 'NSQResponseSchema', 'NSQMessageSchema',
    'NSQErrorSchema', 'FrameType', 'ConnectionStatus', 'NSQCommands'
)



from .base import NSQHTTPConnection
from ..utils import convert_to_str


class NSQDHTTPWriter(NSQHTTPConnection):
    """
    :see: http://nsq.io/components/nsqd.html
    """

    async def ping(self):
        """Monitoring endpoint.
        :returns: should return `"OK"`, otherwise raises an exception.
        """
        return self.perform_request('GET', 'ping', None, None)

    async def info(self):
        """Returns version information."""
        resp = await self.perform_request('GET', 'info', None, None)
        return resp

    async def stats(self):
        """Returns version information."""
        resp = await self.perform_request(
            'GET', 'stats', {'format': 'json'}, None)
        return resp

    async def pub(self, topic, message):
        """Returns version information."""
        resp = await self.perform_request(
            'POST', 'pub', {'topic': topic}, message)
        return resp

    async def mpub(self, topic, *messages):
        """Returns version information."""
        assert len(messages), "Specify one or mor message"
        _msgs = [convert_to_str(m) for m in messages]
        msgs = '\n'.join(_msgs)
        resp = await self.perform_request(
            'POST', 'mpub', {'topic': topic}, msgs)
        return resp

    async def create_topic(self, topic):
        resp = await self.perform_request(
            'POST', 'topic/create', {'topic': topic}, None)
        return resp

    async def delete_topic(self, topic):
        resp = await self.perform_request(
            'POST', 'topic/delete', {'topic': topic}, None)
        return resp

    async def create_channel(self, topic, channel):
        resp = await self.perform_request(
            'POST', 'channel/create', {'topic': topic, 'channel': channel},
            None)
        return resp

    async def delete_channel(self, topic, channel):
        resp = await self.perform_request(
            'POST', 'channel/delete', {'topic': topic, 'channel': channel},
            None)
        return resp

    async def empty_topic(self, topic):
        resp = await self.perform_request(
            'POST', 'topic/empty', {'topic': topic}, None)
        return resp

    async def topic_pause(self, topic):
        resp = await self.perform_request(
            'POST', 'topic/pause', {'topic': topic}, None)
        return resp

    async def topic_unpause(self, topic):
        resp = await self.perform_request(
            'POST', 'topic/unpause', {'topic': topic}, None)
        return resp

    async def pause_channel(self, channel, topic):
        resp = await self.perform_request(
            'POST', 'channel/pause', {'topic': topic, 'channel': channel},
            None)
        return resp

    async def unpause_channel(self, channel, topic):
        resp = await self.perform_request(
            'POST', 'channel/unpause', {'topic': topic, 'channel': channel},
            None)
        return resp

    async def debug_pprof(self):
        resp = await self.perform_request(
            'GET', 'debug/pprof', None, None)
        return resp

    async def debug_pprof_profile(self):
        resp = await self.perform_request(
            'GET', 'debug/pprof/profile', None, None)
        return resp

    async def debug_pprof_goroutine(self):
        resp = await self.perform_request(
            'GET', 'debug/pprof/goroutine', None, None)
        return resp

    async def debug_pprof_heap(self):
        resp = await self.perform_request(
            'GET', 'debug/pprof/heap', None, None)
        return resp

    async def debug_pprof_block(self):
        resp = await self.perform_request(
            'GET', 'debug/pprof/block', None, None)
        return resp

    async def debug_pprof_threadcreate(self):
        resp = await self.perform_request(
            'GET', 'debug/pprof/threadcreate', None, None)
        return resp

    async def nsqlookupd_tcp_addresses(self):
        """
        List of nsqlookupd TCP addresses.
        """
        resp = await self.perform_request(
            'GET', 'config/nsqlookupd_tcp_addresses', None, None)
        return resp



from typing import Union

from . import NSQCommands, FrameType


class NSQResponseSchema:
    """NSQ Response schema"""
    body: bytes
    frame_type: FrameType

    def __init__(self, body: bytes, frame_type: Union[FrameType, int] = None):
        self.body = body
        self.frame_type = (
            frame_type
            if isinstance(frame_type, FrameType) or frame_type is None
            else FrameType(frame_type)
        )

    def __repr__(self):
        return '<NSQResponseSchema frame_type:{}, body:{}, is_ok:{}>'.format(
            self.frame_type, self.body, self.is_ok)

    def __bool__(self):
        return True

    @property
    def is_ok(self) -> bool:
        return self.body == NSQCommands.OK

    @property
    def is_heartbeat(self) -> bool:
        return self.body == b'_heartbeat_'

    @property
    def is_message(self) -> bool:
        return self.frame_type.is_message

    @property
    def is_response(self) -> bool:
        return self.frame_type.is_response

    @property
    def is_error(self) -> bool:
        return self.frame_type.is_error


class NSQMessageSchema(NSQResponseSchema):
    """NSQ Message schema"""
    timestamp: int = None
    attempts: int = None
    id: str = None

    def __init__(
            self, timestamp: int, attempts: int, id_: bytes, body: bytes,
            frame_type: Union[FrameType, int]):
        super().__init__(body, frame_type)
        self.timestamp = timestamp
        self.attempts = attempts
        self.id = id_.decode('utf-8')

    def __repr__(self):
        return (
            '<NSQMessageSchema frame_type:{}, body:{}, timestamp:{}, '
            'attempts:{}, id:{}>'
        ).format(
            self.frame_type, self.body, self.timestamp, self.attempts,
            self.id
        )


class NSQErrorSchema(NSQResponseSchema):
    """NSQ Error"""
    code: str

    def __init__(
            self, code: bytes, body: bytes,
            frame_type: Union[FrameType, int]):
        super().__init__(body, frame_type)
        self.code = code.decode('utf-8')

    def __repr__(self):
        return '<NSQErrorSchema frame_type:{}, body:{}, code:{}>'.format(
            self.frame_type, self.body, self.code)

    def __bool__(self):
        return False



class NSQCommands:
    MAGIC_V2 = b'  V2'
    OK = b'OK'
    BIN_OK = b'\x00\x00\x00\x06\x00\x00\x00\x00OK'
    IDENTIFY = b'IDENTIFY'
    NOP = b'NOP'
    FIN = b'FIN'
    REQ = b'REQ'
    TOUCH = b'TOUCH'
    RDY = b'RDY'
    MPUB = b'MPUB'
    CLS = b'CLS'
    AUTH = b'AUTH'
    SUB = b'SUB'
    PUB = b'PUB'
    DPUB = b'DPUB'



import sys
from setuptools import setup, find_packages
from ansq import __version__

PY_VER = sys.version_info
install_requires = [
    'aiohttp>=1.0.0',
]

classifiers = [
    'License :: OSI Approved :: MIT License',
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
]

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='ansq',
    version=__version__.__version__,
    description='Written with native Asyncio NSQ package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=classifiers,
    author='LiST',
    author_email='support@list.family',
    url='https://github.com/list-family/ansq',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    install_requires=install_requires,
    include_package_data=True,
)



from functools import wraps
from time import time
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ansq.tcp.connection import NSQConnection
    from . import NSQMessageSchema

__all__ = 'NSQMessage'


def not_processed(func):
    """Decorator to verify that the message has not yet been processed.

    :raises RuntimeWarning: in case message was processed earlier.
    """
    @wraps(func)
    async def decorator(cls: 'NSQMessage', *args, **kwargs):
        if cls.is_processed:
            raise RuntimeWarning('Message has already been processed')
        response = await func(cls, *args, **kwargs)
        return response
    return decorator


class NSQMessage:
    def __init__(
            self, message_schema: 'NSQMessageSchema',
            connection: 'NSQConnection',
            timeout: int = 60000, is_processed: bool = False):
        self.timestamp = message_schema.timestamp
        self.attempts = message_schema.attempts
        self.body = message_schema.body
        self.id = message_schema.id
        self._connection = connection
        self._is_processed = is_processed

        self._timeout_in = timeout
        self._initialized_at = time()

    def __repr__(self):
        return (
            '<NSQMessage id="{id}", body={body}, attempts={attempts}, '
            'timestamp={timestamp}, timeout={timeout}, '
            'initialized_at={initialized_at}, is_timed_out={is_timed_out}, '
            'is_processed={is_processed}>'.format(
                id=self.id, body=self.body, attempts=self.attempts,
                timestamp=self.timestamp, timeout=self.timeout,
                initialized_at=self._initialized_at,
                is_timed_out=self.is_timed_out, is_processed=self.is_processed
            )
        )

    def __str__(self):
        """Returns decoded message's body.

        :raises UnicodeDecodeError: Trying to decode bytes like ``b'\xa1'``.
            Be careful. Call this method only if you sure that the body is str.
        """
        return self.body.decode('utf-8')

    @property
    def is_processed(self) -> bool:
        """True if message has been processed:
            * finished
            * re-queued
            * timed out
        """
        return self.is_timed_out or self._is_processed

    @property
    def timeout(self) -> float:
        return self._timeout_in

    @property
    def is_timed_out(self) -> bool:
        return self._initialized_at + self._timeout_in * 0.001 < time()

    @not_processed
    async def fin(self):
        """Finish a message (indicate successful processing)

        :raises RuntimeWarning: in case message was processed earlier.
        """
        await self._connection.fin(self.id)
        self._is_processed = True

    @not_processed
    async def req(self, timeout=10):
        """Re-queue a message (indicate failure to process)

        :param timeout: ``int`` configured max timeout  0 is a special case
            that will not defer re-queueing.
        :raises RuntimeWarning: in case message was processed earlier.
        """
        await self._connection.req(self.id, timeout)
        self._is_processed = True

    @not_processed
    async def touch(self):
        """Reset the timeout for an in-flight message.

        :raises RuntimeWarning: in case message was processed earlier.
        """
        await self._connection.touch(self.id)



from .base import NSQHTTPConnection


class NsqLookupd(NSQHTTPConnection):
    """
    :see: http://nsq.io/components/nsqlookupd.html
    """

    async def ping(self):
        """Monitoring endpoint.
        :returns: should return `"OK"`, otherwise raises an exception.
        """
        return self.perform_request('GET', 'ping', None, None)

    async def info(self):
        """Returns version information."""
        response = await self.perform_request('GET', 'info', None, None)
        return response

    async def lookup(self, topic):
        """XXX

        :param topic:
        :return:
        """
        response = await self.perform_request(
            'GET', 'lookup', {'topic': topic}, None)
        return response

    async def topics(self):
        """XXX

        :return:
        """
        resp = await self.perform_request('GET', 'topics', None, None)
        return resp

    async def channels(self, topic):
        """XXX

        :param topic:
        :return:
        """
        resp = await self.perform_request(
            'GET', 'channels', {'topic': topic}, None)
        return resp

    async def nodes(self):
        """XXX

        :return:
        """
        resp = await self.perform_request('GET', 'nodes', None, None)
        return resp

    async def create_topic(self, topic):
        """XXX

        :param topic:
        :return:
        """
        resp = await self.perform_request(
            'POST', '/topic/create', {'topic': topic}, None)
        return resp

    async def delete_topic(self, topic):
        """XXX

        :param topic:
        :return:
        """
        resp = await self.perform_request(
            'POST', '/topic/delete', {'topic': topic}, None)
        return resp

    async def create_channel(self, topic, channel):
        """XXX

        :param topic:
        :param channel:
        :return:
        """
        resp = await self.perform_request(
            'POST', '/channel/create', {'topic': topic, 'channel': channel},
            None)
        return resp

    async def delete_channel(self, topic, channel):
        """XXX

        :param topic:
        :param channel:
        :return:
        """
        resp = await self.perform_request(
            'POST', '/channel/delete', {'topic': topic, 'channel': channel},
            None)
        return resp

    async def tombstone_topic_producer(self, topic, node):
        """XXX

        :param topic:
        :param node:
        :return:
        """
        resp = await self.perform_request(
            'POST', 'delete_channel', {'topic': topic, 'node': node},
            None)
        return resp



import asyncio
import json
import logging
import sys
from asyncio.events import AbstractEventLoop
from time import time
from typing import Optional, Callable, Any, Union, Generator

from ansq.tcp.exceptions import ProtocolError, get_exception, NSQUnauthorized
from ansq.tcp.types import (
    TCPConnection as NSQConnectionBase, ConnectionStatus, NSQMessage,
    NSQResponseSchema, NSQMessageSchema, NSQErrorSchema, NSQCommands
)
from ansq.utils import validate_topic_channel_name


class NSQConnection(NSQConnectionBase):
    async def connect(self) -> bool:
        """Open connection"""
        self._reader, self._writer = await asyncio.open_connection(
            self._host, self._port)

        self._writer.write(NSQCommands.MAGIC_V2)
        self._status = ConnectionStatus.CONNECTED
        self.logger.info('Connect to {} established'.format(self.endpoint))

        self._reader_task = self._loop.create_task(self._read_data_task())

        return True

    async def reconnect(self, raise_error: bool = True) -> bool:
        """Reconnect method will reopen the connection,
        send the ``identify`` command with your or default config,
        authorize you if you were authorized
        and resubscribe to previous topic/channel.

        So, after that command you can continue to work
        with NSQ like nothing is happened.

        :param raise_error: If ``False``, method will log exception
            and return the ``bool`` value anyway.
        :type raise_error: :class:`bool`

        :returns: Reconnect successful status.
        """
        self.logger.debug('Reconnecting to {}...'.format(self.endpoint))
        self._status = ConnectionStatus.RECONNECTING

        await self._do_close(change_status=False)
        try:
            await self.connect()
            await self.identify()
            self._secret and await self.auth(self._secret)
            self._is_subscribed and await self.subscribe(
                self._topic, self._channel, self.rdy_messages_count)
        except Exception as e:
            if raise_error:
                raise e

            await self._do_close(e)
            return False

        self.logger.info('Reconnected to {}'.format(self.endpoint))
        self._status = ConnectionStatus.CONNECTED
        return True

    async def _do_close(
            self, exception: Exception = None, change_status: bool = True):
        if self.is_closed or self.status.is_init or self.status.is_closing:
            return

        if change_status:
            self._status = ConnectionStatus.CLOSING

        if exception:
            self.logger.error(
                'Connection {} is closing due an error: {}'.format(
                    self.endpoint, exception))
        else:
            self.logger.debug('Connection {} is closing...'.format(
                self.endpoint))

        if self.is_subscribed and change_status:
            self._is_subscribed = False
            while self._message_queue.qsize() > 0:
                self._message_queue.get_nowait()
            self._message_queue.put_nowait(None)

        if self._reader_task and not self._reader_task.done():
            self._reader_task.cancel()
            try:
                await self._reader_task
            except Exception as e:
                self.logger.exception(e)

        if (
                change_status
                and self._reconnect_task
                and not self._reconnect_task.done()
        ):
            self._reconnect_task.cancel()
            try:
                await self._reconnect_task
            except Exception as e:
                self.logger.exception(e)

        try:
            self._writer.close()
            if sys.version_info >= (3, 7):
                await self._writer.wait_closed()
        finally:
            pass

        if self._message_queue.qsize() > 0:
            self._message_queue.get_nowait()

        if change_status:
            self._status = ConnectionStatus.CLOSED
            self.logger.info('Connection {} is closed'.format(self.endpoint))

    async def execute(
            self, command: Union[str, bytes], *args, data: Any = None,
            callback: Callable = None
    ) -> Optional[Union[NSQResponseSchema, NSQErrorSchema, NSQMessageSchema]]:
        """Execute command

        Be careful: commands ``NOP``, ``FIN``, ``RDY``, ``REQ``, ``TOUCH``
            by NSQ spec returns ``None`` as  The class:`asyncio.Future` result.

        :returns: The response from NSQ.
        """
        if command is None:
            raise ValueError('Command must not be None')
        if None in set(args):
            raise ValueError('Args must not contain None')

        if (
                self.is_auth_required
                and not self.is_authorized
                and command != NSQCommands.AUTH
        ):
            raise NSQUnauthorized('NSQ server requires client authorization')

        if self.status.is_reconnecting and not self._reconnect_task:
            await self._reconnect_task

        assert self._reader, 'You should call `connect` method first'
        assert self._status or command == NSQCommands.CLS, (
            'Connection is closed')

        future = self._loop.create_future()
        if command in (
                NSQCommands.NOP, NSQCommands.FIN, NSQCommands.RDY,
                NSQCommands.REQ, NSQCommands.TOUCH
        ):
            future.set_result(None)
            callback and callback(None)
        else:
            self._cmd_waiters.append((future, callback))

        command_raw = self._parser.encode_command(command, *args, data=data)
        self.logger.debug('NSQ: Executing command %s' % command_raw)
        self._writer.write(command_raw)

        # track all processed and requeued messages
        if command in (
                NSQCommands.FIN, NSQCommands.REQ,
                NSQCommands.FIN.decode(), NSQCommands.REQ.decode()
        ):
            self._in_flight = max(0, self._in_flight - 1)

        return await future

    async def identify(self, config: Union[dict, str] = None, **kwargs):
        if config and isinstance(config, (dict, str)):
            raise TypeError('Config should be dict type or str')

        if config or kwargs:
            self._config = config or kwargs
        config = json.dumps(self._config)

        response = await self.execute(
            NSQCommands.IDENTIFY, data=config, callback=self._start_upgrading)

        if response in (NSQCommands.OK, NSQCommands.OK.decode()):
            await self._finish_upgrading()
            return response

        response_config = json.loads(response.body)
        fut = None

        if response_config.get('auth_required'):
            self._is_auth_required = True
        if response_config.get('tls_v1'):
            await self._upgrade_to_tls()
        if response_config.get('snappy'):
            fut = self._upgrade_to_snappy()
        elif response_config.get('deflate'):
            fut = self._upgrade_to_deflate()
        await self._finish_upgrading()

        if fut:
            upgrade_response = await fut
            assert upgrade_response.is_ok

        return response

    async def _upgrade_to_tls(self):
        raise NotImplementedError('Upgrade to TLSv1 not implemented yet')

    def _upgrade_to_snappy(self):
        raise NotImplementedError('Upgrade to snappy not implemented yet')

    def _upgrade_to_deflate(self):
        raise NotImplementedError('Upgrade to deflate not implemented yet')

    async def _read_data_task(self):
        """Response reader task."""
        while not self._reader.at_eof():
            try:
                data = await self._reader.read(52)
            except asyncio.CancelledError:
                # useful during update to TLS, task canceled but connection
                # should not be closed
                return
            except Exception as exc:
                await self._do_close(exc)
                return

            self._parser.feed(data)
            not self._is_upgrading and await self._read_buffer()

        self.logger.info('Lost connection to NSQ')
        if self._auto_reconnect:
            await asyncio.sleep(1)
            self._reconnect_task = self._loop.create_task(
                self.reconnect(raise_error=False)
            )
        else:
            await self._do_close(OSError('Lost connection to NSQ'))

    async def _parse_data(self) -> bool:
        try:
            response = self._parser.get()
        except ProtocolError as exc:
            # ProtocolError is fatal
            await self._do_close(exc)
            return False

        if response is None:
            return False

        self.logger.debug('NSQ: Got data: %s', response)

        if response.is_heartbeat:
            await self._pulse()
            return True

        if response.is_message:
            # track number in flight messages
            self._in_flight += 1
            self._on_message_hook(response)
            return True

        future, callback = self._cmd_waiters.popleft()
        future: asyncio.Future
        callback: Callable

        if response.is_response:
            if not future.cancelled():
                future.set_result(response)
                callback is not None and callback(response)

        if response.is_error:
            exception = get_exception(response.code, response.body)

            if not future.cancelled():
                future.set_result(response)
            callback and callback(response)
            self._on_exception and self._on_exception(exception)

        return True

    def _on_message_hook(self, message_schema: NSQMessageSchema):
        self._last_message_timestamp = time()
        message = NSQMessage(message_schema, self)

        if self._on_message:
            try:
                message = self._on_message(message)
            except Exception as e:
                self._do_close(e)
        self._message_queue.put_nowait(message)

    async def _read_buffer(self):
        is_continue = True
        while is_continue:
            is_continue = await self._parse_data()

    def _start_upgrading(self, resp=None):
        self._is_upgrading = True

    async def _finish_upgrading(self, resp=None):
        await self._read_buffer()
        self._is_upgrading = False

    async def auth(self, secret: str) -> Union[
        NSQResponseSchema, NSQErrorSchema
    ]:
        """If the ``IDENTIFY`` response indicates ``auth_required=true``
        the client must send ``AUTH`` before any ``SUB``, ``PUB`` or ``MPUB``
        commands. If auth_required is not present (or ``false``),
        a client must not authorize.

        :param secret:
        :return:
        """
        response = await self.execute(NSQCommands.AUTH, data=secret)
        if not isinstance(response, NSQErrorSchema):
            self._secret = secret
        return response

    async def sub(self, topic, channel) -> Union[
        NSQResponseSchema, NSQErrorSchema
    ]:
        """Subscribe to the topic and channel"""
        validate_topic_channel_name(topic)
        validate_topic_channel_name(channel)
        response = await self.execute(NSQCommands.SUB, topic, channel)
        if isinstance(response, NSQResponseSchema):
            self._topic = topic
            self._channel = channel
            self._is_subscribed = True
        return response

    async def pub(self, topic, message) -> Union[
        NSQResponseSchema, NSQErrorSchema
    ]:
        """Publish a message to a topic"""
        validate_topic_channel_name(topic)
        return await self.execute(
            NSQCommands.PUB, topic, data=message)

    async def dpub(self, topic, message, delay_time) -> Union[
        NSQResponseSchema, NSQErrorSchema
    ]:
        """Publish a deferred message to a topic"""
        validate_topic_channel_name(topic)
        return await self.execute(
            NSQCommands.DPUB, topic, delay_time, data=message)

    async def mpub(self, topic, *messages) -> Union[
        NSQResponseSchema, NSQErrorSchema
    ]:
        """Publish multiple messages to a topic"""
        validate_topic_channel_name(topic)
        return await self.execute(
            NSQCommands.MPUB, topic,
            data=messages if len(messages) > 1 else messages[0]
        )

    async def rdy(self, messages_count: int = 1):
        """Update RDY state (indicate you are ready to receive N messages)"""
        assert isinstance(messages_count, int), (
            'Argument messages_count should be positive integer')
        assert messages_count >= 0, (
            'Argument messages_count should be positive integer')

        self.rdy_messages_count = messages_count
        await self.execute(NSQCommands.RDY, messages_count)

    async def fin(self, message_id: Union[str, NSQMessage]):
        """Finish a message (indicate successful processing)"""
        if isinstance(message_id, NSQMessage):
            await message_id.fin()
        await self.execute(NSQCommands.FIN, message_id)

    async def req(self, message_id: Union[str, NSQMessage], timeout: int = 0):
        """Re-queue a message (indicate failure to process)

        The re-queued message is placed at the tail of the queue,
        equivalent to having just published it.
        """
        if isinstance(message_id, NSQMessage):
            await message_id.req(timeout)
        await self.execute(NSQCommands.REQ, message_id, timeout)

    async def touch(self, message_id: Union[str, NSQMessage]):
        """Reset the timeout for an in-flight message"""
        if isinstance(message_id, NSQMessage):
            await message_id.touch()
        await self.execute(NSQCommands.TOUCH, message_id)

    async def _cls(self) -> Union[NSQResponseSchema, NSQErrorSchema]:
        return await self.execute(NSQCommands.CLS)

    async def subscribe(
            self, topic: str, channel: str, messages_count: int = 1):
        """Shortcut for ``sub()`` and ``rdy()`` methods"""
        sub_response = await self.sub(topic, channel)
        if not sub_response:
            return sub_response

        await self.rdy(messages_count)

    async def messages(self) -> Generator[NSQMessage, None, None]:
        """Generator, yields messages"""
        assert self.is_subscribed, (
            'You should subscribe to the topic first')

        while self.is_subscribed:
            message = await self._message_queue.get()
            if message is None:
                return
            yield message

    def get_message(self) -> Optional[NSQMessage]:
        """Shortcut for ``asyncio.Queue.get_nowait()``
        without raising exceptions
        """
        try:
            return self._message_queue.get_nowait()
        except asyncio.QueueEmpty:
            return None

    async def wait_for_message(self) -> NSQMessage:
        """Shortcut for `asyncio.Queue.get()``.

        :rtype: :class:`NSQMessage`
        :returns: :class:`NSQMessage`.
            Be aware on closing action may returns ``None``.
            This is need to exit from ``NSQConnection.messages`` generator
            when connection closed with exception.
        """
        return await self.message_queue.get()


async def open_connection(
        host: str = 'localhost', port: int = 4150, *,
        message_queue: asyncio.Queue = None, on_message: Callable = None,
        on_exception: Callable = None, loop: AbstractEventLoop = None,
        auto_reconnect: bool = True, heartbeat_interval: int = 30000,
        feature_negotiation: bool = True, tls_v1: bool = False,
        snappy: bool = False, deflate: bool = False,
        deflate_level: int = 6, sample_rate: int = 0,
        debug: bool = False, logger: logging.Logger = None
) -> NSQConnection:
    nsq = NSQConnection(
        host, port, message_queue=message_queue, on_message=on_message,
        on_exception=on_exception, loop=loop, auto_reconnect=auto_reconnect,
        heartbeat_interval=heartbeat_interval,
        feature_negotiation=feature_negotiation, tls_v1=tls_v1,
        snappy=snappy, deflate=deflate, deflate_level=deflate_level,
        sample_rate=sample_rate, debug=debug, logger=logger
    )
    await nsq.connect()
    await nsq.identify()
    return nsq
